from django.test import TestCase
from django.utils import timezone
from eveuniverse.models import EveType

from app_utils.testing import create_user_from_evecharacter

from skillfarm.models.skillfarm import (
    CharacterSkill,
    CharacterSkillqueueEntry,
    SkillFarmAudit,
    SkillFarmSetup,
)
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import create_skillfarm_character

MODULE_PATH = "skillfarm.models.skillfarm"


class TestSkillfarmModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_skillfarm_character(1001)

    def test_should_return_string_audit(self):
        """Test should return the Audit Character Data"""
        self.assertEqual(str(self.audit), "Gneuten's Character Data")

    def test_should_return_esi_scopes(self):
        """Test should return the ESI Scopes for Skillfarm"""
        self.assertEqual(
            self.audit.get_esi_scopes(),
            ["esi-skills.read_skills.v1", "esi-skills.read_skillqueue.v1"],
        )

    def test_is_active_should_return_false(self):
        """Test should return False for is_active Property"""
        self.assertFalse(self.audit.is_active)

    def test_is_active_should_return_true(self):
        """Test should return True for is_active Property"""
        self.audit.last_update_skills = timezone.now()
        self.audit.last_update_skillqueue = timezone.now()

        self.assertTrue(self.audit.is_active)

    def test_is_active_should_update_active_to_false(self):
        """Test should update the active field to False"""
        self.audit.active = True
        self.audit.last_update_skills = timezone.now() - timezone.timedelta(days=7)
        self.audit.last_update_skillqueue = timezone.now() - timezone.timedelta(days=7)
        self.audit.save()

        self.assertFalse(self.audit.is_active)

    def test_is_cooldown_should_return_false(self):
        """Test should return False for is_cooldown Property"""
        self.assertFalse(self.audit.is_cooldown)

    def test_is_cooldown_should_return_true(self):
        """Test should return True for is_cooldown Property"""
        self.audit.last_notification = timezone.now()
        self.assertTrue(self.audit.is_cooldown)


class TestSkillFarmHelper(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.user_ownership = create_user_from_evecharacter(character_id=1001)

        cls.audit = SkillFarmAudit(
            name=cls.user_ownership.character.character_name,
            character=cls.user_ownership.character,
            last_update_skillqueue=timezone.now(),
            last_update_skills=timezone.now(),
        )
        cls.audit.save()

        cls.skill1 = EveType.objects.get(name="skill1")
        cls.skill2 = EveType.objects.get(name="skill2")

    def test_should_return_skill_extraction_no_skillset(self):
        """Test should return the Skill Extraction with empty list"""
        self.assertEqual(self.audit.skill_extraction(), [])

    def test_should_return_skill_extraction_skillset_without_level5(self):
        """Test should return the Skill Extraction with empty list"""
        skillfarmsetup = SkillFarmSetup.objects.create(
            character=self.audit, skillset=[self.skill1.name, self.skill2.name]
        )
        skillfarmsetup.save()

        CharacterSkill.objects.create(
            character=self.audit,
            eve_type=self.skill1,
            active_skill_level=3,
            skillpoints_in_skill=1000,
            trained_skill_level=4,
        )

        self.assertEqual(self.audit.skill_extraction(), [])

    def test_should_return_skill_extraction_skillset_with_level5(self):
        """Test should return the Skill Extraction with skill names"""
        skillfarmsetup = SkillFarmSetup.objects.create(
            character=self.audit, skillset=[self.skill1.name, self.skill2.name]
        )
        skillfarmsetup.save()

        CharacterSkill.objects.create(
            character=self.audit,
            eve_type=self.skill1,
            active_skill_level=4,
            skillpoints_in_skill=1000,
            trained_skill_level=5,
        )

        self.assertEqual(self.audit.skill_extraction(), ["skill1"])

    def test_should_return_skillqueue_extraction_empty(self):
        """Test should return the Skillqueue Extraction with empty list"""
        self.assertEqual(self.audit.skillqueue_extraction(), [])

    def test_should_return_skillqueue_extraction_skillset_with_finished_skill(self):
        """Test should return the Skillqueue Extraction with skill names"""
        skillfarmsetup = SkillFarmSetup.objects.create(
            character=self.audit, skillset=[self.skill1.name, self.skill2.name]
        )
        skillfarmsetup.save()

        CharacterSkillqueueEntry.objects.create(
            queue_position=1,
            eve_type=self.skill2,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=20),
            finish_date=timezone.now() - timezone.timedelta(days=3),
            character=self.audit,
        )

        self.assertEqual(self.audit.skillqueue_extraction(), ["skill2"])

    def test_should_return_skillqueue_extraction_skillset_without_finished_skill(self):
        """Test should return the Skillqueue Extraction with skill names"""
        skillfarmsetup = SkillFarmSetup.objects.create(
            character=self.audit, skillset=[self.skill1.name, self.skill2.name]
        )
        skillfarmsetup.save()

        CharacterSkillqueueEntry.objects.create(
            queue_position=1,
            eve_type=self.skill2,
            finished_level=4,
            start_date=timezone.now() - timezone.timedelta(days=20),
            finish_date=timezone.now() - timezone.timedelta(days=3),
            character=self.audit,
        )

        self.assertEqual(self.audit.skillqueue_extraction(), [])

    def test_get_finished_skills_should_return_empty(self):
        """Test should return the finished skills with empty list"""
        self.assertEqual(self.audit.get_finished_skills(), [])

    def test_get_finished_skills_should_return_skills(self):
        """Test should return the finished skills with skill names"""
        skillfarmsetup = SkillFarmSetup.objects.create(
            character=self.audit, skillset=[self.skill1.name, self.skill2.name]
        )
        skillfarmsetup.save()

        CharacterSkill.objects.create(
            character=self.audit,
            eve_type=self.skill1,
            active_skill_level=5,
            skillpoints_in_skill=1000,
            trained_skill_level=5,
        )

        CharacterSkillqueueEntry.objects.create(
            queue_position=1,
            eve_type=self.skill2,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=20),
            finish_date=timezone.now() - timezone.timedelta(days=3),
            character=self.audit,
        )

        self.assertEqual(sorted(self.audit.get_finished_skills()), ["skill1", "skill2"])



from django.db import models, transaction
from eveuniverse.models import EveType

from skillfarm.hooks import get_extension_logger
from skillfarm.providers import esi
from skillfarm.task_helper import (
    HTTPGatewayTimeoutError,
    NotModifiedError,
    etag_results,
)

logger = get_extension_logger(__name__)


class SkillqueueManager(models.Manager):
    def update_or_create_esi(self, character, force_refresh: bool = False):
        """Update or create skills queue for a character from ESI."""
        skillqueue = self._fetch_data_from_esi(character, force_refresh=force_refresh)

        if skillqueue is None:
            return False

        entries = []

        for entry in skillqueue:
            eve_type_instance, _ = EveType.objects.get_or_create_esi(
                id=entry.get("skill_id")
            )
            entries.append(
                self.model(
                    name=character.name,
                    character=character,
                    eve_type=eve_type_instance,
                    finish_date=entry.get("finish_date"),
                    finished_level=entry.get("finished_level"),
                    level_end_sp=entry.get("level_end_sp"),
                    level_start_sp=entry.get("level_start_sp"),
                    queue_position=entry.get("queue_position"),
                    start_date=entry.get("start_date"),
                    training_start_sp=entry.get("training_start_sp"),
                )
            )

        self._atomic_write(character, entries)
        return True

    def _fetch_data_from_esi(
        self, character, force_refresh: bool = False
    ) -> list[dict]:
        logger.debug("%s: Fetching skill queue from ESI", character)

        skillqueue = None
        token = character.get_token()
        try:
            skillqueue_data = esi.client.Skills.get_characters_character_id_skillqueue(
                character_id=character.character.character_id,
            )

            skillqueue = etag_results(
                skillqueue_data, token, force_refresh=force_refresh
            )
        except NotModifiedError:
            logger.debug(
                "No New Skillqueue data for: %s", character.character.character_name
            )
        except HTTPGatewayTimeoutError:
            # TODO Add retry logic?
            logger.debug(
                "Skillqueue data ESI Timeout for: %s",
                character.character.character_name,
            )

        return skillqueue

    @transaction.atomic()
    def _atomic_write(self, character, entries):
        self.filter(character=character).delete()

        if not entries:
            logger.info("%s: Skill queue is empty", character)
            return

        self.bulk_create(entries)
        logger.info("%s: Updated %s skill queue/s", character, len(entries))



from skillfarm.api.helpers import arabic_number_to_roman, get_skillset
from skillfarm.models.skillfarm import CharacterSkill, SkillFarmAudit


def _get_character_skills(character: SkillFarmAudit) -> dict:
    """Get all Skills for the current character"""
    skillset = get_skillset(character)
    skills_dict = []
    extraction_ready = False

    if skillset is not None:
        skills = CharacterSkill.objects.filter(
            character=character, eve_type__name__in=skillset
        ).select_related("eve_type")

        for skill in skills:
            if skill.active_skill_level == 0:
                continue
            level = arabic_number_to_roman(skill.active_skill_level)

            dict_data = {
                "skill": f"{skill.eve_type.name} {level}",
                "level": skill.active_skill_level,
                "skillpoints": skill.skillpoints_in_skill,
            }

            if skill.is_exc_ready:
                extraction_ready = True

            skills_dict.append(dict_data)

    output = {
        "skills": skills_dict,
        "is_extraction_ready": extraction_ready,
    }

    return output



from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from allianceauth.eveonline.models import EveCharacter

from skillfarm.hooks import get_extension_logger
from skillfarm.models.skillfarm import SkillFarmAudit, SkillFarmSetup

logger = get_extension_logger(__name__)


def arabic_number_to_roman(value) -> str:
    """Map to convert arabic to roman numbers (1 to 5 only)"""
    my_map = {0: "-", 1: "I", 2: "II", 3: "III", 4: "IV", 5: "V"}
    try:
        return my_map[value]
    except KeyError:
        return "-"


def get_main_character(request, character_id):
    """Get Character and check permissions"""
    perms = True
    try:
        main_char = EveCharacter.objects.get(character_id=character_id)
    except EveCharacter.DoesNotExist:
        main_char = EveCharacter.objects.select_related(
            "character_ownership",
            "character_ownership__user__profile",
            "character_ownership__user__profile__main_character",
        ).get(character_id=request.user.profile.main_character.character_id)

    # check access
    visible = SkillFarmAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        perms = False
    return perms, main_char


def get_character(request, character_id):
    """Get Character and check permissions"""
    perms = True
    try:
        character = SkillFarmAudit.objects.get(character__character_id=character_id)
    except SkillFarmAudit.DoesNotExist:
        return False, None

    # check access
    visible = SkillFarmAudit.objects.visible_to(request.user)
    if character not in visible:
        perms = False
    return perms, character


def get_alts_queryset(main_char, corporations=None):
    """Get all alts for a main character, optionally filtered by corporations."""
    try:
        linked_corporations = main_char.character_ownership.user.character_ownerships.all().select_related(
            "character_ownership"
        )

        if corporations:
            linked_corporations = linked_corporations.filter(
                character__corporation_id__in=corporations
            )

        linked_corporations = linked_corporations.values_list("character_id", flat=True)

        return EveCharacter.objects.filter(id__in=linked_corporations)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_skillset(character: SkillFarmAudit) -> dict | None:
    """Get the skillset for the character"""
    try:
        skillfilter = SkillFarmSetup.objects.get(character=character)
        return skillfilter.skillset
    except SkillFarmSetup.DoesNotExist:
        return None


def generate_button(template, queryset, settings, request) -> mark_safe:
    """Generate a html button for the tax system"""
    return format_html(
        render_to_string(
            template,
            {
                "queryset": queryset,
                "settings": settings,
            },
            request=request,
        )
    )


# pylint: disable=too-many-positional-arguments
def generate_settings(
    title: str, icon: str, color: str, text: str, modal: str, action: str, ajax: str
) -> dict:
    """Generate a settings dict for the tax system"""
    return {
        "title": title,
        "icon": icon,
        "color": color,
        "text": text,
        "modal": modal,
        "action": action,
        "ajax": ajax,
    }


def generate_progressbar(progress) -> str:
    """Generate a progress bar"""
    value = round(progress, 2)
    if value > 50:
        progress_value = format_html('<span class="text-white)">{}%</span>', value)
    else:
        progress_value = format_html('<span class="text-dark">{}%</span>', value)

    progressbar = format_html(
        """
        <div class="progress-outer flex-grow-1 me-2">
            <div class="progress" style="position: relative;">
                <div class="progress-bar progress-bar-warning progress-bar-striped active" role="progressbar" style="width: {}%; box-shadow: -1px 3px 5px rgba(0, 180, 231, 0.9);"></div>
                <div class="fw-bold fs-6 text-center position-absolute top-50 start-50 translate-middle">{}</div>
            </div>
        </div>
        """,
        value,
        progress_value,
    )

    return progressbar



from django.contrib.auth.models import User
from eveuniverse.models import EveType

from allianceauth.authentication.backends import StateBackend
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user

from skillfarm.models.prices import EveTypePrice
from skillfarm.models.skillfarm import CharacterSkill, SkillFarmAudit, SkillFarmSetup


def create_character(eve_character: EveCharacter, **kwargs) -> SkillFarmAudit:
    """Create a Skillfarm Character from EveCharacter"""
    params = {"name": eve_character.character_name, "character": eve_character}
    params.update(kwargs)
    character = SkillFarmAudit(**params)
    character.save()
    return character


def create_user_from_evecharacter_with_access(
    character_id: int, disconnect_signals: bool = True
) -> tuple[User, CharacterOwnership]:
    """Create user with access from an existing eve character and use it as main."""
    auth_character = EveCharacter.objects.get(character_id=character_id)
    username = StateBackend.iterate_username(auth_character.character_name)
    user = AuthUtils.create_user(username, disconnect_signals=disconnect_signals)
    user = AuthUtils.add_permission_to_user_by_name(
        "skillfarm.basic_access", user, disconnect_signals=disconnect_signals
    )
    character_ownership = add_character_to_user(
        user,
        auth_character,
        is_main=True,
        scopes=SkillFarmAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )
    return user, character_ownership


def create_skillfarm_character(character_id: int, **kwargs) -> SkillFarmAudit:
    """Create a Audit Character from a existing EveCharacter"""

    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=True
    )
    return create_character(character_ownership.character, **kwargs)


def create_skillsetup_character(character_id: int, skillset: list) -> CharacterSkill:
    """Create a SkillSet for Skillfarm Audit Character"""
    audit = SkillFarmAudit.objects.get(
        character__character_id=character_id,
    )

    skillset = SkillFarmSetup.objects.create(
        character=audit,
        skillset=skillset,
    )

    return skillset


def create_evetypeprice(evetype_id: int, **kwargs) -> EveType:
    params = {
        "eve_type": EveType.objects.get(id=evetype_id),
    }
    params.update(kwargs)
    price = EveTypePrice(**params)
    price.save()
    return price


def create_skill_character(
    character_id: int,
    evetype_id: int,
    skillpoints: int,
    active_level: int = 5,
    trained_level: int = 5,
) -> CharacterSkill:
    """Create a Skill for Skillfarm Audit Character"""
    audit = SkillFarmAudit.objects.get(
        character__character_id=character_id,
    )

    skill = CharacterSkill.objects.create(
        character=audit,
        eve_type=EveType.objects.get(id=evetype_id),
        skillpoints_in_skill=skillpoints,
        active_skill_level=active_level,
        trained_skill_level=trained_level,
    )

    return skill


def add_auth_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True
) -> CharacterOwnership:
    auth_character = EveCharacter.objects.get(character_id=character_id)
    return add_character_to_user(
        user,
        auth_character,
        is_main=False,
        scopes=SkillFarmAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )


def add_skillfarmaudit_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> CharacterOwnership:
    character_ownership = add_auth_character_to_user(
        user,
        character_id,
        disconnect_signals=disconnect_signals,
    )
    return create_character(character_ownership.character, **kwargs)



from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from skillfarm.api.helpers import (
    generate_button,
    generate_progressbar,
    generate_settings,
)
from skillfarm.hooks import get_extension_logger
from skillfarm.models.skillfarm import CharacterSkillqueueEntry, SkillFarmAudit

logger = get_extension_logger(__name__)


def _calculate_single_progress_bar(skill: CharacterSkillqueueEntry):
    """Calculate the progress bar for a single skill"""
    totalsp = skill.level_end_sp
    start_date = skill.start_date
    finish_date = skill.finish_date

    if totalsp == 0:
        return 0

    current_date = timezone.now()
    total_duration = (finish_date - start_date).total_seconds()
    elapsed_duration = (current_date - start_date).total_seconds()

    if elapsed_duration > total_duration:
        progress = 100
    else:
        progress = (elapsed_duration / total_duration) * 100

    # Ensure the progress percentage is between 0 and 100
    progress = max(progress, 0)

    return round(progress, 2)


def _calculate_sum_progress_bar(skillqueue: dict):
    """Calculate the progress bar for the skillqueue"""
    # Calculate the progress percentage for each skill individually
    total_progress_percent = 0
    skill_count = len(skillqueue)

    if skill_count == 0:
        return generate_progressbar(0)

    for skill in skillqueue:
        if skill["start_date"] and skill["finish_date"] == "-":
            continue
        total_progress_percent += skill["progress"]["value"]

    # Calculate the average progress percentage
    average_progress_percent = total_progress_percent / skill_count

    return generate_progressbar(average_progress_percent)


# pylint: disable=too-many-arguments, too-many-positional-arguments
def generate_action_settings(
    character: SkillFarmAudit,
    title,
    icon,
    color,
    modal,
    viewname,
    request,
    text: str = "",
):
    """Generate a settings dict for the skillfarm"""
    url = reverse(
        viewname=viewname,
        kwargs={
            "character_id": character.character.character_id,
        },
    )
    settings = generate_settings(
        title=title,
        icon=icon,
        color=color,
        text=text,
        modal=modal,
        action=url,
        ajax="action",
    )
    return generate_button(
        "skillfarm/partials/forms/button.html",
        character,
        settings,
        request,
    )


def _skillfarm_actions(character: SkillFarmAudit, perms, request):
    """Generate the skillfarm actions buttons for Skill Details"""
    actions = []
    if perms:
        actions.append(
            generate_action_settings(
                character=character,
                title=_("Toggle Alarm"),
                icon="fas fa-bullhorn",
                color="primary",
                modal="skillfarm-confirm",
                viewname="skillfarm:switch_alarm",
                request=request,
                text=_("Are you sure you want to toggle the alarm for {}?").format(
                    character.character.character_name
                ),
            )
        )
        actions.append(
            generate_action_settings(
                character=character,
                title=_("Edit Skillset"),
                icon="fas fa-pencil",
                color="warning",
                modal="skillfarm-skillset",
                viewname="skillfarm:skillset",
                request=request,
            )
        )
        actions.append(
            generate_action_settings(
                character=character,
                title=_("Delete Character"),
                icon="fas fa-trash",
                color="danger",
                modal="skillfarm-confirm",
                viewname="skillfarm:delete_character",
                request=request,
                text=_("Are you sure you want to delete {}?").format(
                    character.character.character_name
                ),
            )
        )

    actions_html = format_html("".join(actions))
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)


def _get_skillinfo_actions(character: SkillFarmAudit, request):
    """Get the skillinfo actions for the skillfarm"""
    settings = generate_settings(
        title=_("Skillinfo"),
        icon="fas fa-info",
        color="primary",
        text="",
        modal="modalViewSkillContainer",
        action=reverse(
            viewname="skillfarm:api:get_skillinfo_details",
            kwargs={
                "character_id": character.character.character_id,
            },
        ),
        ajax="ajax_skillview",
    )
    return generate_button(
        "skillfarm/partials/forms/button.html",
        character,
        settings,
        request,
    )


def _get_extraction_icon(skills, skillqueue) -> str:
    """Get the extraction icon"""
    if skills is True:
        title = _("Skill Extraction Ready")
        div_id = "skillfarm-skill-extractor"
        icon = "fas fa-exclamation-triangle"
        color = "red"
    elif skillqueue is True:
        title = _("Please check your Character a Skill should be ready for extraction")
        div_id = "skillfarm-skill-extractor-maybe"
        icon = "fas fa-question"
        color = "orange"
    else:
        return ""
    return (
        "<img src='/static/skillfarm/images/skillExtractor.png' data-tooltip-toggle='skillfarm-tooltip' class='rounded-circle' id='"
        + div_id
        + "'style='width: 32px'>"
        + f"<i class='{icon}' style='margin-left: 5px; color: {color}' title='"
        + title
        + "' data-tooltip-toggle='skillfarm-tooltip'></i>"
    )


def _get_notification_icon(status: bool) -> str:
    """Get the notification icon"""
    html = "<i class='fa-solid fa-bullhorn' style='margin-left: 5px; color:"
    if status:
        html += (
            "green' title='"
            + _("Notification Activated")
            + "' data-tooltip-toggle='skillfarm-tooltip'></i>"
        )
    else:
        html += (
            "gray' title='"
            + _("Notification Deactivated")
            + "' data-tooltip-toggle='skillfarm-tooltip'></i>"
        )
    return html



from django.contrib import admin
from django.utils.html import format_html

from allianceauth.eveonline.evelinks import eveimageserver

from skillfarm.models.skillfarm import SkillFarmAudit


# TODO make a ETAG Cache clear option?
@admin.register(SkillFarmAudit)
class SkillFarmAuditAdmin(admin.ModelAdmin):
    """Admin Interface for Characters"""

    model = SkillFarmAudit
    model._meta.verbose_name = "Character"
    model._meta.verbose_name_plural = "Characters"

    list_display = (
        "_entity_pic",
        "_character__character_id",
        "_character__character_name",
        "_last_update_skillqueue",
        "_last_update_skills",
    )

    list_display_links = (
        "_entity_pic",
        "_character__character_id",
        "_character__character_name",
    )

    list_select_related = ("character",)

    ordering = ["character__character_name"]

    search_fields = ["character__character_name", "character__character_id"]

    actions = [
        "delete_objects",
    ]

    @admin.display(description="")
    def _entity_pic(self, obj: SkillFarmAudit):
        eve_id = obj.character.character_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("character", eve_id, 32),
        )

    @admin.display(description="Character ID", ordering="character__character_id")
    def _character__character_id(self, obj: SkillFarmAudit):
        return obj.character.character_id

    @admin.display(description="Character Name", ordering="character__character_name")
    def _character__character_name(self, obj: SkillFarmAudit):
        return obj.character.character_name

    @admin.display(
        description="Last Update Skillqueue", ordering="last_update_skillqueue"
    )
    def _last_update_skillqueue(self, obj: SkillFarmAudit):
        return obj.last_update_skillqueue

    @admin.display(description="Last Update Skills", ordering="last_update_skills")
    def _last_update_skills(self, obj: SkillFarmAudit):
        return obj.last_update_skills

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False



"""App Tasks"""

import datetime

import requests

# Third Party
from celery import chain, shared_task

from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import Error
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify
from allianceauth.services.tasks import QueueOnce

from skillfarm import app_settings
from skillfarm.decorators import when_esi_is_available
from skillfarm.hooks import get_extension_logger
from skillfarm.models.prices import EveTypePrice
from skillfarm.models.skillfarm import (
    CharacterSkill,
    CharacterSkillqueueEntry,
    SkillFarmAudit,
)

logger = get_extension_logger(__name__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.SKILLFARM_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

_update_skillfarm_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["character_id"], "graceful": True}},
}


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_skillfarm(runs: int = 0):
    characters = SkillFarmAudit.objects.select_related("character").all()
    for character in characters:
        update_character_skillfarm.apply_async(args=[character.character.character_id])
        runs = runs + 1
    logger.info("Queued %s Skillfarm Updates", runs)


@shared_task(**_update_skillfarm_params)
@when_esi_is_available
def update_character_skillfarm(
    character_id, force_refresh=False
):  # pylint: disable=unused-argument
    character = SkillFarmAudit.objects.get(character__character_id=character_id)

    # Settings for the Task Queue
    que = []
    skip_date = timezone.now() - datetime.timedelta(
        hours=app_settings.SKILLFARM_STALE_STATUS
    )
    mindt = timezone.now() - datetime.timedelta(days=7)
    priority = 7

    logger.debug(
        "Processing Audit Updates for %s", format(character.character.character_name)
    )
    if (character.last_update_skillqueue or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_skillqueue.si(character_id, force_refresh=force_refresh).set(
                priority=priority
            )
        )

    if (character.last_update_skills or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_skills.si(character_id, force_refresh=force_refresh).set(
                priority=priority
            )
        )

    chain(que).apply_async()
    logger.debug("Queued %s Tasks for %s", len(que), character.character.character_name)


@shared_task(**_update_skillfarm_params)
def update_char_skillqueue(character_id, force_refresh=False):
    character = SkillFarmAudit.objects.get(character__character_id=character_id)
    CharacterSkillqueueEntry.objects.update_or_create_esi(
        character, force_refresh=force_refresh
    )
    character.last_update_skillqueue = timezone.now()
    character.save()


@shared_task(**_update_skillfarm_params)
def update_char_skills(character_id, force_refresh=False):
    character = SkillFarmAudit.objects.get(character__character_id=character_id)
    CharacterSkill.objects.update_or_create_esi(character, force_refresh=force_refresh)
    character.last_update_skills = timezone.now()
    character.save()


@shared_task(**TASK_DEFAULTS_ONCE)
def check_skillfarm_notifications(runs: int = 0):
    characters = SkillFarmAudit.objects.filter(active=True)
    notified_characters = []

    # Create a dictionary to map main characters to their alts
    main_to_alts = {}
    for character in characters:
        try:
            main_character = (
                character.character.character_ownership.user.profile.main_character
            )
        except ObjectDoesNotExist:
            logger.warning(
                "Main Character not found for %s, skipping notification",
                character.character.character_name,
            )
            continue

        if main_character not in main_to_alts:
            main_to_alts[main_character] = []
        main_to_alts[main_character].append(character)

    for main_character, alts in main_to_alts.items():
        msg_items = []
        for alt in alts:
            if alt.notification:
                skill_names = alt.get_finished_skills()
                if skill_names:
                    # Create and Add Notification Message
                    msg = alt._generate_notification(skill_names)
                    msg_items.append(msg)
                    notified_characters.append(alt)
            else:
                # Reset Settings for Alts that have no notification enabled
                alt.notification_sent = False
                alt.last_notification = None
                alt.save()

        if msg_items:
            # Add each message to Main Character
            notifiy_message = "\n".join(msg_items)
            logger.debug(
                "Skilltraining has been finished for %s Skills: %s",
                main_character.character_name,
                main_character,
            )
            title = _("Skillfarm Notifications")
            full_message = format_html(
                "Following Skills have finished training: \n{}", notifiy_message
            )

            notify(
                title=title,
                message=full_message,
                user=main_character.character_ownership.user,
                level="warning",
            )
            runs = runs + 1

    if notified_characters:
        # Set notification_sent to True for all characters that were notified
        for character in notified_characters:
            character.notification_sent = True
            character.last_notification = timezone.now()
            character.save()

    logger.info("Queued %s Skillfarm Notifications", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
def update_all_prices():
    prices = EveTypePrice.objects.all()
    market_data = {}

    if len(prices) == 0:
        logger.info("No Prices to update")
        return

    request = requests.get(
        "https://market.fuzzwork.co.uk/aggregates/",
        params={
            "types": ",".join([str(x) for x in prices]),
            "station": app_settings.SKILLFARM_PRICE_SOURCE_ID,
        },
    ).json()

    market_data.update(request)

    for price in prices:
        if price.eve_type.id in market_data:
            price.buy = float(market_data[price.eve_type.id]["buy"]["percentile"])
            price.sell = float(market_data[price.eve_type.id]["sell"]["percentile"])
            price.updated_at = timezone.now()

    try:
        EveTypePrice.objects.bulk_update(prices, ["buy", "sell", "updated_at"])
    except Error as e:
        logger.error("Error updating prices: %s", e)
        return

    logger.info("Skillfarm Prices updated")
